/* -*- c++ -*- */
/* 
 * Copyright 2014 SSL, VNU-UET.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "sync_end_node_impl.h"
#include <stdexcept>
#include <iostream>
#include <string.h>
#include <cstdio>
#include <complex.h>
#include <math.h>
#include <time.h>

namespace gr {
  namespace PNC {

    sync_end_node::sptr
    sync_end_node::make()
    {
      return gnuradio::get_initial_sptr
        (new sync_end_node_impl());
    }

    /*
     * The private constructor
     */
    sync_end_node_impl::sync_end_node_impl()
      : gr::block("sync_end_node",
              gr::io_signature::make(1, 1, sizeof(gr_complex)),
              gr::io_signature::make2(2, 2, sizeof(gr_complex), sizeof(int))),
	      d_window_size(16), d_window_index(0), d_xcorr_index(0), d_sts_size(16), d_threshold(3.0), 
	      d_nPeaks_detected(0), d_check_full_xcorr(false), d_shift_count(0), d_lts_sym_count(0), d_beacon_index(0),
	      d_beacon_count(0), d_beacon_frame_size(40), d_plot_index(0), d_check_full_plot(false),
	      d_state(ST_IDLE)
    {
	/* initialize window to plot xcorr */
	std::vector<double> d_xcorr_plot(450, 0.0);

	/* initialize beacon frame */
	std::vector<int> d_beacon_frame(d_beacon_frame_size, 0);

	/* initialize header frame */
	std::vector<gr_complex> d_header(d_beacon_frame_size, gr_complex(0.0, 0.0));

	/* initialize one sts symbol */
	d_sts = generate_sts(d_window_size);

	/* initialize cross-correlation window */
	d_xcorr = initialize_xcorr_window(d_window_size);

	/* initialize observation window */
	d_window = initialize_obs_window(d_window_size);
    }

    /*
     * Our virtual destructor.
     */
    sync_end_node_impl::~sync_end_node_impl()
    {
    }

    void
    sync_end_node_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
        /* <+forecast+> e.g. ninput_items_required[0] = noutput_items */
    }

    /* initialize cross-correlation window */
    std::vector<double> 
    initialize_xcorr_window(int d_window_size)
    {
	std::vector<double> xcorr_window(d_window_size, 0.0);
	return xcorr_window;
    }

    /* initialize observation window */
    std::vector<gr_complex>
    initialize_obs_window(int d_window_size)
    {
	std::vector<gr_complex> window(d_window_size, gr_complex(0.0, 0.0));	
	return window;
    }

    /* initialize one short training symbol */
    std::vector<gr_complex>
    generate_sts(int d_window_size)
    {
	std::vector<gr_complex> sq_sts(d_window_size, gr_complex(0, 0));
	sq_sts[0]  = gr_complex( 0.2041, 0.2041); /* short training symbol 1   */
	sq_sts[1]  = gr_complex(-1.1755, 0.0208); /* short training symbol 2   */
	sq_sts[2]  = gr_complex(-0.1196,-0.6969); /* short training symbol 3   */
	sq_sts[3]  = gr_complex( 1.2670,-0.1123); /* short training symbol 4   */
	sq_sts[4]  = gr_complex( 0.8165, 0.0000); /* short training symbol 5   */
	sq_sts[5]  = gr_complex( 1.2670,-0.1123); /* short training symbol 6   */
	sq_sts[6]  = gr_complex(-0.1196,-0.6969); /* short training symbol 7   */
	sq_sts[7]  = gr_complex(-1.1755, 0.0208); /* short training symbol 8   */
	sq_sts[8]  = gr_complex( 0.4082, 0.4082); /* short training symbol 9   */
	sq_sts[9]  = gr_complex( 0.0208,-1.1755); /* short training symbol 10  */
	sq_sts[10] = gr_complex(-0.6969,-0.1196); /* short training symbol 11  */
	sq_sts[11] = gr_complex(-0.1123, 1.2670); /* short training symbol 12  */
	sq_sts[12] = gr_complex( 0.0000, 0.8165); /* short training symbol 13  */
	sq_sts[13] = gr_complex(-0.1123, 1.2670); /* short training symbol 14  */
	sq_sts[14] = gr_complex(-0.6969,-0.1196); /* short training symbol 15  */
	sq_sts[15] = gr_complex( 0.0208,-1.1755); /* short training symbol 16  */
	return sq_sts;
    }

    /* shift observation window function */
    std::vector<gr_complex>
    shift_window_func(std::vector<gr_complex> obs_window, gr_complex new_sym, int d_window_size)
    {
	for (int i = 0; i<d_window_size-1; i++)
	{
		obs_window[i] = obs_window[i+1];
	}
	obs_window[d_window_size-1] = new_sym;
	return obs_window;
    }

    /*shift cross-correlation window function */
    std::vector<double>
    shift_xcorr_func(std::vector<double> xcorr, double new_xcorr, int d_window_size)
    {
	for (int i = 0; i<d_window_size-1; i++)
	{
		xcorr[i] = xcorr[i+1];
	}
	xcorr[d_window_size-1] = new_xcorr;
	return xcorr;
    }

    /* start general work */
    int
    sync_end_node_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
	int nInputItems = ninput_items[0];

	int nOutputItems = noutput_items;

        const gr_complex *in = (const gr_complex *) input_items[0];

        gr_complex *out0 = (gr_complex *) output_items[0];

        int *out1 = (int *) output_items[1];

	int ni = 0;
	int no = 0;

	while (ni < nInputItems && no < nOutputItems)
	{
		switch (d_state)
		{
			case ST_IDLE:
			{
				/* We need to wait until receiving enough symbols */
				if(ni<d_sts_size-1)
				{
					d_window[d_window_index] = in[ni];
					d_window_index++;
					ni++;
				}
				else
				{
					d_window[d_window_index] = in[ni];
					ni++;
					d_state = ST_WAIT_STS_SYM;					
				}
				break;						
			}
			case ST_WAIT_STS_SYM:
			{
				d_window = shift_window_func(d_window, in[ni], d_window_size); /* shift observation window */
				/* start calculating cross-correlation */
				gr_complex temp_xcorr = gr_complex(0, 0);
				double normalization_coeff = 0.0;
				for(int i = 0; i<d_sts_size; i++)
				{
					temp_xcorr = temp_xcorr + conj(d_sts[i])*d_window[i];
					normalization_coeff = normalization_coeff + abs(conj(d_window[i])*d_window[i]);
				}
				double temp_xcorr_norm = abs(temp_xcorr)/normalization_coeff;
				/*-------------------------------*/
				if (d_check_full_plot==false)
				{
					d_xcorr_plot[d_plot_index] = temp_xcorr_norm;
					d_plot_index++;
					int c = 0;
					c = d_plot_index;
					if(c==450) d_check_full_plot = true;
				}
				else
				{
					for(int f = 0; f<449; f++)
					{
						d_xcorr_plot[f] = d_xcorr_plot[f+1];
					}
					d_xcorr_plot[450] = temp_xcorr_norm;
				}
				/*-------------------------------*/
				if(d_check_full_xcorr==false)
				{
					d_xcorr[d_xcorr_index] = temp_xcorr_norm;
					if (d_xcorr_index<(d_window_size-1))					
					{
						d_xcorr_index++;
						ni++;
						break;
					}
					else
					{
						d_check_full_xcorr = true;
						ni++;
						break;						
					}
				}
				/* start finding the first peak (out of ten peaks) in the cross-correlation window */
				if(d_check_full_xcorr==true)
				{
					d_xcorr = shift_xcorr_func(d_xcorr, temp_xcorr_norm, d_window_size);
					bool check_max_curr = true;
					for(int j = 0; j<(d_window_size-1); j++)
					{
						if (d_xcorr[j]>d_xcorr[d_xcorr_index]) 
						{ 
							check_max_curr = false;
							break; 
						}
					}
					if(check_max_curr==false) 
					{ 
						ni++; 
						break; 
					}
					else
					{
						double mean_xcorr = 0.0;
						double temp_sum = 0.0;
						for(int k = 0; k<d_window_size; k++)
						{
							temp_sum = temp_sum + d_xcorr[k];
						}
						mean_xcorr = temp_sum/d_window_size;
						if((d_xcorr[d_xcorr_index]/mean_xcorr)>=d_threshold)
						{
							d_nPeaks_detected++;  /* number of peaks = 1 */
							d_state = ST_STS_SYM; /* the first peak detected */
							ni++;
							break;
						}
					}
					
				}
				break;	
			}
			case ST_STS_SYM:
			{		
				d_shift_count++;				
				d_window = shift_window_func(d_window, in[ni], d_window_size); /* shift observation window */
				/* start calculating cross-correlation */
				gr_complex temp_xcorr = gr_complex(0, 0);
				double normalization_coeff = 0;
				for(int i = 0; i<d_sts_size; i++)
				{
					temp_xcorr = temp_xcorr + conj(d_sts[i])*d_window[i];
					normalization_coeff = normalization_coeff + abs(conj(d_window[i])*d_window[i]);
				}
				double temp_xcorr_norm = abs(temp_xcorr)/normalization_coeff;
				/*-------------------------------*/
				if (d_check_full_plot==false)
				{
					d_xcorr_plot[d_plot_index] = temp_xcorr_norm;
					d_plot_index++;
					int c = 0;
					c = d_plot_index;
					if(c==450) d_check_full_plot = true;
				}
				else
				{
					for(int f = 0; f<449; f++)
					{
						d_xcorr_plot[f] = d_xcorr_plot[f+1];
					}
					d_xcorr_plot[450] = temp_xcorr_norm;
				}
				/*-------------------------------*/
				
				d_xcorr = shift_xcorr_func(d_xcorr, temp_xcorr_norm, d_window_size);
				bool check_max_curr = true;
				for(int j = 0; j<(d_window_size-1); j++)
				{
					if (d_xcorr[j]>d_xcorr[d_xcorr_index]) 
					{ 
						check_max_curr = false;
						break; 
					}
				}
				if(check_max_curr==false) 
				{
					if(d_shift_count==d_window_size)
					{
						d_shift_count = 0;
						d_nPeaks_detected = 0;
						d_state = ST_WAIT_STS_SYM; /* back to ST_WAIT_STS_SYM */
					}	 
					ni++;  
					break; 
				}
				else
				{
					double mean_xcorr = 0.0;
					double temp_sum = 0.0;
					for(int j = 0; j<d_window_size; j++)
					{
						temp_sum = temp_sum + d_xcorr[j];
					}
					mean_xcorr = temp_sum/d_window_size;
					if((d_xcorr[d_xcorr_index]/mean_xcorr)>=d_threshold)
					{
						if(d_shift_count<d_window_size)
						{
							d_shift_count = 0;
							d_nPeaks_detected = 1;
							ni++;
							break;
						}
						if(d_shift_count==d_window_size)						
						{
							d_nPeaks_detected++;
							d_shift_count = 0;
							if(d_nPeaks_detected==10) /* the last peak detected, this sample is the end of STS */
							{
								d_state = ST_LTS_SYM;
							}
							ni++;
							break;
						}
					}
				}
				break;
			}
			case ST_LTS_SYM:
			{
				/* carrier frequency offset can be estimated here by using the two lts symbols */
				/*-------------------------------*/
				d_window = shift_window_func(d_window, in[ni], d_window_size); /* shift observation window */
				gr_complex temp_xcorr = gr_complex(0, 0);
				double normalization_coeff = 0;
				for(int i = 0; i<d_sts_size; i++)
				{
					temp_xcorr = temp_xcorr + conj(d_sts[i])*d_window[i];
					normalization_coeff = normalization_coeff + abs(conj(d_window[i])*d_window[i]);
				}
				double temp_xcorr_norm = abs(temp_xcorr)/normalization_coeff;
				if (d_check_full_plot==false)
				{
					d_xcorr_plot[d_plot_index] = temp_xcorr_norm;
					d_plot_index++;
					int c = 0;
					c = d_plot_index;
					if(c==450) d_check_full_plot = true;
				}
				else
				{
					for(int f = 0; f<449; f++)
					{
						d_xcorr_plot[f] = d_xcorr_plot[f+1];
					}
					d_xcorr_plot[450] = temp_xcorr_norm;
				}
				/*-------------------------------*/

				d_lts_sym_count++;
				if (d_lts_sym_count==160)
				{
					d_state = ST_CHECK_IF_BEACON_OR_DATA;
				}
				ni++;
				break;
			}
			case ST_CHECK_IF_BEACON_OR_DATA:
			{
				/*check the first 5 bytes equivalent to 40 samples */
				/*-------------------------------*/
				d_window = shift_window_func(d_window, in[ni], d_window_size); /* shift observation window */
				gr_complex temp_xcorr = gr_complex(0, 0);
				double normalization_coeff = 0;
				for(int i = 0; i<d_sts_size; i++)
				{
					temp_xcorr = temp_xcorr + conj(d_sts[i])*d_window[i];
					normalization_coeff = normalization_coeff + abs(conj(d_window[i])*d_window[i]);
				}
				double temp_xcorr_norm = abs(temp_xcorr)/normalization_coeff;
				if (d_check_full_plot==false)
				{
					d_xcorr_plot[d_plot_index] = temp_xcorr_norm;
					d_plot_index++;
					int c = 0;
					c = d_plot_index;
					if(c==450) d_check_full_plot = true;
				}
				else
				{
					for(int f = 0; f<449; f++)
					{
						d_xcorr_plot[f] = d_xcorr_plot[f+1];
					}
					d_xcorr_plot[450] = temp_xcorr_norm;
				}
				/*-------------------------------*/

				d_beacon_count++;
				gr_complex sym = in[ni];
				if (sym.real()>0)
				{
					d_beacon_frame[d_beacon_index] = 1;
					d_header[d_beacon_index] = sym;
				}
				if (d_beacon_count==d_beacon_frame_size)
				{
					int sum = 0;
					for(int i = 0; i<d_beacon_frame_size; i++)
					{
						sum = sum + d_beacon_frame[i];
					}
					if (sum>=35)
					{
						d_state = ST_BEACON_DETECTED;
						ni++;
						break;
					}
					else 
					{
						d_state = ST_DATA_DETECTED;
						ni++;
						break;
					}
				}
				d_beacon_index++;
				ni++;
				break;
			}
			case ST_BEACON_DETECTED:
			{
				/* pass a beacon signal to the controller block */
				std::cout<<"Beacon detected :-)"<<std::endl;
				for (int i = 0; i<d_beacon_frame_size; i++)
				{
					std::cout<<d_beacon_frame[i]<<" ";
				}
				std::cout<<std::endl;
				/*-------------------------------------------*/
				std::cout<<"Cross-correlation: "<<std::endl;
				for (int j = 0; j<450; j++)
				{
					std::cout<<d_xcorr_plot[j]<<" ";
				}
				std::cout<<std::endl;
				/*-------------------------------------------*/
				out1[no] = 1; /* pass a signal to tell the controller to send data*/
				ni++;				
				no++;
				d_state = ST_SLEEP;
				break;
			}
			case ST_DATA_DETECTED:
			{
				/* pass data to the next signal processing block*/
				
				ni++;
				break;
			}
			case ST_SLEEP:
			{
				/*do nothing*/
				ni++;
				break;
			}	
		}	
	}

        // Tell runtime system how many input items we consumed on
        // each input stream.
        consume_each (ni);

        // Tell runtime system how many output items we produced.
        return no;
    }

  } /* namespace PNC */
} /* namespace gr */

